package controller;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.LinkedList;
import java.util.Stack;

import javax.swing.DefaultComboBoxModel;
import javax.swing.JComboBox;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import javax.swing.event.MouseInputListener;

import controller.listener.AntiAliasedLineListener;
import controller.listener.BezierCurveListener;
import controller.listener.CircleListener;
import controller.listener.ClippingRectangleListener;
import controller.listener.LineListener;
import controller.listener.PointListener;
import controller.listener.PolygonFilledListener;
import controller.listener.PolygonListener;
import controller.listener.RectangleListener;
import controller.listener.TransformListener;

import model.Matrix;
import model.drawables.DrawableObject;
import model.drawables.Point;
import model.drawables.Polygon;
import model.drawables.Rectangle;

import view.DrawingPanelView;

/**
 * Der Controller, der eine DrawingPanelView verwaltet und das dazugehörige
 * Model in Form einer Liste von DrawableObjects. Das Model des Controllers ist
 * hart mit dem der View verbunden. Änderungen werden also direkt in der View
 * passieren. Ggf. muss die View aufgefordert werden sich neu zu zeichnen.
 * 
 * @author Nicolas Neubauer
 * 
 */
public class DrawingPanelViewController implements DrawableObjectProcessing {

	// View
	private DrawingPanelView drawingPanelView;

	// Model
	private LinkedList<DrawableObject> drawableObjectsModel;
	private DrawableObject temporaryObject;
	private Stack<Matrix> matrixStackModel;

	// Listener für die einzelnen grafischen Objekte
	private PointListener pointListener;
	private LineListener lineListener;
	private AntiAliasedLineListener antiAliasedLineListener;
	private RectangleListener rectangleListener;
	private CircleListener circleListener;
	private PolygonListener polygonListener;
	private PolygonFilledListener polygonFilledListener;
	private ClippingRectangleListener clippingRectangleListener;
	private TransformListener transformListener;
	private BezierCurveListener bezierCurveListener;

	/**
	 * Der Konstruktor initialisiert die View und legt die Listener an.
	 */
	public DrawingPanelViewController() {

		drawableObjectsModel = new LinkedList<DrawableObject>();
		matrixStackModel = new Stack<Matrix>();
		drawingPanelView = new DrawingPanelView(640, 480, drawableObjectsModel);

		// das UI anpassen
		drawingPanelView.getButton().setText("Clear");

		// welche Objekte sollen gezeichnet werden können?
		String[] objectArray = { "Punkt", "Linie", "Anti-Aliased Linie",
				"Rechteck", "Polygon", "gefuelltes Polygon", "Kreis",
				"Polygon-Clipping", "Polygon-Transformation", "Bezier-Kurve" };
		drawingPanelView.getComboBox().setModel(
				new DefaultComboBoxModel(objectArray));

		// Listener für die einzelnen Grafischen Objekte anlegen.
		pointListener = new PointListener(this);
		lineListener = new LineListener(this);
		antiAliasedLineListener = new AntiAliasedLineListener(this);
		rectangleListener = new RectangleListener(this);
		circleListener = new CircleListener(this);
		polygonListener = new PolygonListener(this);
		polygonFilledListener = new PolygonFilledListener(this);
		clippingRectangleListener = new ClippingRectangleListener(this);
		transformListener = new TransformListener(this);
		bezierCurveListener = new BezierCurveListener(this);

		// Event-Listener fuer Button
		ActionListener a = new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				// Das Model leeren und die View neu zeichnen
				drawableObjectsModel.clear();
				drawingPanelView.getDrawingPanel().repaint();
			}

		};

		// ChangeListener um festzustellen, was gezeichnet werden soll
		ActionListener c = new ActionListener() {

			/**
			 * Hängt den passenden Listener an die Zeichenfläche
			 */
			public void actionPerformed(ActionEvent e) {
				// Hänge das aktuell gewählte Objekt als Listener an die
				// View

				JComboBox cb = (JComboBox) e.getSource();

				if (cb.getSelectedItem().equals("Punkt"))
					changeMouseInputListenerTo(pointListener);
				else if (cb.getSelectedItem().equals("Linie"))
					changeMouseInputListenerTo(lineListener);
				else if (cb.getSelectedItem().equals("Anti-Aliased Linie"))
					changeMouseInputListenerTo(antiAliasedLineListener);
				else if (cb.getSelectedItem().equals("Rechteck"))
					changeMouseInputListenerTo(rectangleListener);
				else if (cb.getSelectedItem().equals("Kreis"))
					changeMouseInputListenerTo(circleListener);
				else if (cb.getSelectedItem().equals("Polygon"))
					changeMouseInputListenerTo(polygonListener);
				else if (cb.getSelectedItem().equals("gefuelltes Polygon"))
					changeMouseInputListenerTo(polygonFilledListener);
				else if (cb.getSelectedItem().equals("Polygon-Clipping"))
					changeMouseInputListenerTo(clippingRectangleListener);
				else if (cb.getSelectedItem().equals("Polygon-Transformation"))
					changeMouseInputListenerTo(transformListener);
				else if (cb.getSelectedItem().equals("Bezier-Kurve"))
					changeMouseInputListenerTo(bezierCurveListener);

			}

			/**
			 * Convenience-Methode zum umsetzen des aktuellen MouseListeners an
			 * der Zeichenfläche der verwalteten View
			 * 
			 * @param newListener
			 *            der zu setzende Listener, alle anderen werden
			 *            ausgehängt
			 */
			private void changeMouseInputListenerTo(
					MouseInputListener newListener) {

				// InputListeners
				MouseListener[] currentInputListeners = drawingPanelView
						.getDrawingPanel().getMouseListeners();

				for (MouseListener curListener : currentInputListeners)
					drawingPanelView.getDrawingPanel().removeMouseListener(
							curListener);

				// MotionListeners
				MouseMotionListener[] currentMotionListners = drawingPanelView
						.getDrawingPanel().getMouseMotionListeners();
				for (MouseMotionListener curListener : currentMotionListners)
					drawingPanelView.getDrawingPanel()
							.removeMouseMotionListener(curListener);

				drawingPanelView.getDrawingPanel()
						.addMouseListener(newListener);
				drawingPanelView.getDrawingPanel().addMouseMotionListener(
						newListener);
			}

		};

		// ActionListener um die Matrix auf den Stack zu pushen
		ActionListener addMatrix = new ActionListener() {

			public void actionPerformed(ActionEvent e) {

				// Lokales Array erstellen
				double[][] matrix = new double[3][3];

				// Versuche die Textfelder als Doubles zu interpretieren
				try {

					// Durch alle Textfelder gehen und Doubles rausziehen
					for (int row = 0; row < 3; row++) {
						for (int col = 0; col < 3; col++) {
							matrix[row][col] = Double
									.parseDouble(drawingPanelView
											.getMatrixPanel().getTextFields()[row][col]
											.getText());
						}
					}

				} catch (NumberFormatException ex) {
					JOptionPane
							.showMessageDialog(
									SwingUtilities
											.getWindowAncestor(drawingPanelView),
									"Eine oder mehrere Zahl(en) ist/sind fehlerhaft formatiert.",
									"Fehler", JOptionPane.ERROR_MESSAGE);
					return; // Keine weitere Verarbeitung
				}

				// Matrix erstellen und dem Stack hinzufügen
				Matrix m = new Matrix(matrix);
				int index = drawingPanelView.getMatrixPanel()
						.getSelectedIndex();
				if (index != -1) {
					matrixStackModel.add(index + 1, m);
				} else {
					matrixStackModel.push(m);
				}

				// View aktualisieren
				drawingPanelView.getMatrixPanel().getMatrixList().setListData(
						matrixStackModel);

			}
		};

		// ActionListener um die Matrix vom Stack zu popen
		ActionListener popMatrix = new ActionListener() {

			public void actionPerformed(ActionEvent e) {

				if (matrixStackModel.size() > 0) {
					matrixStackModel.remove(drawingPanelView.getMatrixPanel()
							.getSelectedIndex());
					// View aktualisieren
					drawingPanelView.getMatrixPanel().getMatrixList()
							.setListData(matrixStackModel);
				}

			}
		};

		// ActionListener um den Stack zu leeren
		ActionListener clearMatrices = new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				matrixStackModel.clear();
				// View aktualisieren
				drawingPanelView.getMatrixPanel().getMatrixList().setListData(
						matrixStackModel);
			}

		};

		// Anhängen der Listener
		drawingPanelView.getButton().addActionListener(a);
		drawingPanelView.getComboBox().addActionListener(c);

		drawingPanelView.getMatrixPanel().getAddButton().addActionListener(
				addMatrix);
		drawingPanelView.getMatrixPanel().getPopButton().addActionListener(
				popMatrix);
		drawingPanelView.getMatrixPanel().getClearButton().addActionListener(
				clearMatrices);

		// Event einmal auslösen, damit der korrekte Listener angehängt wird
		drawingPanelView.getComboBox().setSelectedItem("Punkt");

	}

	/**
	 * Gibt die verwaltete View zurück
	 * 
	 * @return die View
	 */
	public DrawingPanelView getView() {
		return drawingPanelView;
	}

	/**
	 * Fügt das DrawableObject in das Model ein und fordert die View zum
	 * Neuzeichnen auf.
	 * 
	 * @param drawableObject
	 */
	public void processDrawableObject(DrawableObject drawableObject) {

		drawableObjectsModel.add(drawableObject);
		drawingPanelView.getDrawingPanel().repaint();

	}

	/**
	 * Löscht das temporäre DrawableObject aus dem Model und fordert die View
	 * zum Neuzeichnen auf.
	 */
	public void clearTemporaryDrawableObject() {

		drawableObjectsModel.remove(temporaryObject);
		drawingPanelView.getDrawingPanel().repaint();
		temporaryObject = null;

	}

	/**
	 * Fügt ein temporäres Objekt in das Model ein und fordert die View zum
	 * Neuzeichnen auf. Das Objekt kann mit clearTemporaryDrawableObject wieder
	 * aus dem Model entfernt werden.
	 * 
	 * @param drawableObject
	 */
	public void setTemporaryDrawableObject(DrawableObject drawableObject) {

		// Erst löschen
		if (temporaryObject != null)
			this.clearTemporaryDrawableObject();

		temporaryObject = drawableObject;
		this.processDrawableObject(temporaryObject);

	}

	/**
	 * Clippt alle im Model verwalteten Polygone mit deren clip()-Methode und
	 * entfernt ggf. nicht mehr sinnvoll definierte Polygone aus dem Model.
	 * Initiiert außerdem das Neu-Zeichnen der Zeichenfläche
	 * 
	 * @param r
	 */
	public void clipDrawableObjects(Rectangle r) {
		boolean stillPolygon;
		LinkedList<DrawableObject> toBeDeleted = new LinkedList<DrawableObject>();

		for (DrawableObject d : drawableObjectsModel) {
			if (d instanceof Polygon) {
				Polygon p = (Polygon) d;
				stillPolygon = p.clip(r);
				if (!stillPolygon)
					toBeDeleted.add(p);
			}
		}

		drawableObjectsModel.removeAll(toBeDeleted);
		drawingPanelView.getDrawingPanel().repaint();
	}

	/**
	 * Transformiert alle im Model verwalteten Polygone mit deren
	 * transformBy-Methode, wenn der übergebene Punkt innerhalb des Polygons
	 * liegt, also isNear() true zurück gibt. Aktualisiert anschließend die
	 * View.
	 * 
	 * @param p
	 */
	public void transformDrawableObjectsNear(Point p) {

		Matrix transformMatrix = getCombinedMatrixFromModel();

		for (DrawableObject d : drawableObjectsModel) {
			if (d instanceof Polygon) {
				Polygon polygon = (Polygon) d;
				if (polygon.isNear(p))
					polygon.transformBy(transformMatrix);
			}
		}

		drawingPanelView.getDrawingPanel().repaint();
	}

	/**
	 * Convenience-Methode, um aus dem Stack von Matrizen die Produktmatrix zu
	 * berechnen. Es wird die oberste von links mit der zweiten multipliziert
	 * usw.
	 * 
	 * @return die Produktmatrix des Stacks
	 */
	@SuppressWarnings("unchecked")
	private Matrix getCombinedMatrixFromModel() {
		Matrix returnMatrix = Matrix.createIdentity(3);

		Stack<Matrix> modelCopy = (Stack<Matrix>) matrixStackModel.clone();

		while (!modelCopy.empty())
			returnMatrix = returnMatrix.multiply(modelCopy.pop());

		return returnMatrix;
	}

}
