package model;

import model.drawables.Point;

/**
 * Repräsentiert eine nxm-Matrix
 * 
 * @author Nicolas Neubauer
 * 
 */
public class Matrix {

	private double[][] elements;

	/**
	 * Übernimmt die Referenz auf das 2-Dimensionale Array in Form von
	 * array[zeilen][spalten].
	 * 
	 * @param elements
	 *            die Elemente der Matrix
	 */
	public Matrix(double[][] elements) {
		this.elements = elements;
	}

	/**
	 * Erzeugt eine Einheitsmatrix mit rows Zeilen und Spalten
	 * 
	 * @param rows
	 *            Anzahl der Zeilen/Spalten
	 * @return eine neue Einheitsmatrix
	 */
	public static Matrix createIdentity(int rows) {
		double[][] identity = new double[rows][rows];

		for (int i = 0; i < rows; i++)
			identity[i][i] = 1;

		return new Matrix(identity);
	}

	/**
	 * Gibt die Matrix als String zurück.
	 * 
	 * @return String Representation
	 */
	public String toString() {
		String r = "{";
		// Durch alle Textfelder gehen und Doubles rausziehen
		for (int row = 0; row < elements.length; row++) {
			r += "{";
			for (int col = 0; col < elements[row].length; col++) {
				r += (col + 1 == elements[row].length) ? elements[row][col]
						: elements[row][col] + ", ";
			}
			r += (row + 1) == elements.length ? "}" : "}, \n";
		}

		return r + "}";
	}

	/**
	 * @return Anzahl der Zeilen
	 */
	public int getRows() {
		return elements.length;
	}

	/**
	 * @return Anzahl der Spalten
	 */
	public int getCols() {
		return elements[0].length;
	}

	/**
	 * Der Wert der Matrix an der Stelle (row,col)
	 * 
	 * @param row
	 *            Zeile
	 * @param col
	 *            Spalte
	 * @return Wert in der Matrix
	 */
	public double getValue(int row, int col) {
		return elements[row][col];
	}

	/**
	 * Multipliziert diese Matrix von links mit der übergebenen Matrix m und
	 * gibt das Ergebnis als neue Matrix zurück. (seiteneffektfrei)
	 * 
	 * this * m = ergebnismatrix
	 * 
	 * @param m
	 *            die Matrix mit der diese multipliziert werden soll.
	 * @return die Ergebnismatrix
	 */
	public Matrix multiply(Matrix m) {
		// Wenn �bergebene Matrix nicht null und wenn Zeilen dieser Matrix und
		// Spalten von m �bereinstimmen
		if ((m != null) && (getCols() == m.getRows())) {
			// Heraus kommt eine Matrix, dessen Zeilen = Zeilen der aktuellen
			// Matrix und dessen Spalten = Spalten von m
			double[][] newElements = new double[getRows()][m.getCols()];
			// Neue Matrix = 0 setzen
			for (int a = 0; a < getRows(); a++) {
				for (int b = 0; b < m.getCols(); b++) {
					newElements[a][b] = 0.0;
				}
			}
			// F�r jede Zeile der aktuellen Matrix
			for (int i = 0; i < getRows(); i++) {
				// Durchlaufe jede Spalte von m
				for (int j = 0; j < m.getCols(); j++) {
					// Durchlaufe jede Zeile von m/Spalte der Aktuellen
					for (int k = 0; k < getCols(); k++) {
						// Setze Element der neuen Matrix auf Wert der Summe
						// der Produkte
						newElements[i][j] += elements[i][k] * m.elements[k][j];
					}
				}
			}
			return new Matrix(newElements);
		}
		return null;
	}

	/**
	 * Multipliziert diese Matrix von links mit dem übergebenen Punkt. Dazu
	 * muss der Punkt unter Umständen als 3-elementiger Vector aufgefasst
	 * werden, wobei die homogene Koordinate 1.0 in der dritten Komponente
	 * ergänzt werden kann.
	 * 
	 * @param p
	 *            der Punkt
	 * @return das Ergebnis dieser Matrix * den Punkt
	 */
	public Point multiply(Point p) {
		// Neues 2d-Array mit 3x1 Elementen
		double[][] elements = new double[getRows()][1];
		// Setze erste Zeile auf X-Koordinate des Punktes
		elements[0][0] = p.x;
		// Setze zweite Zeile auf Y-Koordinate des Punktes
		elements[1][0] = p.y;
		// Setze restliche Zeilen auf 1
		for (int i = 2; i < getRows(); i++) {
			elements[i][0] = 1;
		}
		// Neuer 3x1-Vektor
		Matrix mi = new Matrix(elements);
		// Multipliziere die Matrix von links an den neuen
		// Vektor
		mi = this.multiply(mi);
		return new Point((int) (mi.getValue(0, 0) + 0.5), (int) (mi.getValue(1,
				0) + 0.5));
	}

}
