package model.drawables;

import java.awt.Graphics;
import java.util.ArrayList;

/**
 * Ein grafisches Objekt, dass eine Bezierkurve repraesentiert.
 * 
 * @author Denis Meyer
 * @author Niels Meyering
 */
public class BezierCurve extends DrawableObject {

	// Punktliste
	private ArrayList<Point> punkte;
	private int rekursionen;

	/**
	 * Konstruktor
	 * 
	 * @param points
	 *            Punktliste
	 * @param rekursionen
	 */
	public BezierCurve(ArrayList<Point> points, int rekursionen) {
		punkte = new ArrayList<Point>(points.size());
		if ((points != null) && !points.isEmpty()) {
			for (int i = 0; i < points.size(); i++) {
				this.punkte.add(points.get(i));
			}
		} else {
			points = new ArrayList<Point>();
		}
		this.rekursionen = rekursionen;
	}

	/**
	 * Zeichnet Bezierkurve mit engegebenen Stuetzpunkten
	 * 
	 * @param g
	 *            Graphics
	 */
	public void paint(Graphics g) {
		if (g != null) {
			if (punkte.size() >= 4) {
				bezier(punkte.get(0), punkte.get(1), punkte.get(2), punkte
						.get(3), 0, g);
			}
		}
	}

	/**
	 * Berechnet die Bezierkurve rekursiv
	 * 
	 * @param sp0
	 *            1. Sttzpunkt
	 * @param sp1
	 *            2. Sttzpunkt
	 * @param sp2
	 *            3. Sttzpunkt
	 * @param sp3
	 *            4. Sttzpunkt
	 * @param currRek
	 *            Aktuelle Rekursionsanzahl (beginne mit 0)
	 * @param g
	 */
	private void bezier(Point sp0, Point sp1, Point sp2, Point sp3,
			int currRek, Graphics g) {
		Line l;
		// Wenn gewollte Rekursionstiefe erreicht
		if (currRek > rekursionen) {
			// Zeichne eine Linie (+ 0.5, da casten zum int double-Zahl
			// "abschneidet")
			l = new Line(new Point((int) (sp0.x + 0.5), (int) (sp0.y + 0.5)),
					new Point((int) (sp3.x + 0.5), (int) (sp3.y + 0.5)));
			l.paint(g);
		} else {
			// Zhle aktuelle Rekursiontiefe hoch
			currRek++;
			// Halbiere die Strecken
			// SSP12 = Sttzpunkt 1 zu Sttzpunkt 2
			Point ssp12 = new Line(sp0, sp1).getHalf();
			// SSP23 = Sttzpunkt 2 zu Sttzpunkt 3
			Point ssp23 = new Line(sp1, sp2).getHalf();
			// SSP34 = Sttzpunkt 3 zu Sttzpunkt 4
			Point ssp34 = new Line(sp2, sp3).getHalf();
			// Halbiere die Hlften der Strecken (1/4)
			// SSSP1223 = SSP12 zu SSP23
			Point sssp1223 = new Line(ssp12, ssp23).getHalf();
			// SSSP2334 = SSP23 zu SSP34
			Point sssp2334 = new Line(ssp23, ssp34).getHalf();
			// Halbiere die Hlften der Hlften der Strecken (1/8)
			// SSSS12232334 = SSSP1223 zu SSSP2334
			Point ssssp12232334 = new Line(sssp1223, sssp2334).getHalf();

			// Starte die Rekursion
			// Linke Seite
			bezier(sp0, ssp12, sssp1223, ssssp12232334, currRek, g);
			// Rechte Seite
			bezier(ssssp12232334, sssp2334, ssp34, sp3, currRek, g);
		}
	}
}
