package model.drawables;

import java.awt.Graphics;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Collections;

import model.Matrix;

/**
 * Ein grafisches Objekt, dass ein Polygon repraesentiert. Die einzelnen
 * Polygonpunkte sind in einer Liste abgelegt. Der erste Punkt ist nicht doppelt
 * abgelegt, d.h. die letzte Kante ist vom letzten Punkt zum ersten Punkt zu
 * ziehen.
 * 
 * @author Nicolas Neubauer
 */
public class Polygon extends DrawableObject {

	protected List<Point> points; // Liste mit Eckpunkten
	private boolean filled;

	/**
	 * Erstellt ein Polygon
	 * 
	 * @param points
	 *            Punkte-Liste
	 * @param filled
	 *            soll das Polygon gefüllt sein?
	 */
	public Polygon(List<Point> points, boolean filled) {
		this.points = points;
		this.filled = filled;
	}

	/**
	 * Zeichnet das Polygon in den uebergebenen grafischen Kontext und fuellt es
	 * gegebenenfalls.
	 * 
	 * @param g
	 *            der grafische Kontext, in den dieses Objekt sich zeichnen soll
	 */
	public void paint(Graphics g) {
		Line l;
		Point prev = null;

		for (Point p : points) {
			if (prev != null) {
				l = new Line(prev, p);
				l.paint(g);
			}
			prev = p;
		}

		// Polygon schließen, Linie zeichnen
		l = new Line(prev, points.get(0));
		l.paint(g);

		if (filled) {
			scanlineFill(g);
		}
	}

	/**
	 * Gibt true zurück, wenn der Punkt im Polygon liegt, false sonst.
	 * 
	 * @param p
	 *            der zu prüfende Punkt
	 * @return true, wenn p innerhalb des Polygons liegt
	 */
	public boolean isNear(Point p) {
		boolean innen = false; // 0 Schnittpunkte
		Point prev = null;
		boolean pOberP1Unter, pUnterP1Ober;
		int schnittpkt;

		for (Point p1 : points) {
			if (prev != null) {
				// Wenn prev oberhalb und point unterhalb
				pOberP1Unter = (prev.y >= p.y) && (p1.y < p.y);
				// Wenn prev unterhalb und point oberhalb
				pUnterP1Ober = (prev.y < p.y) && (p1.y >= p.y);

				// Wenn nicht Strahl auf einer kompletten Seite des Polygons
				// liegt, heit, wenn der Strahl nicht komplett ber der
				// aktuellen Kante liegt
				if (pOberP1Unter || pUnterP1Ober) {
					// Berechne Schnittpunkt von Kante und Strahl
					schnittpkt = (new Line(prev, p1).intersection(p.y, TOP)).x;

					// Wenn "echter" Schnittpunkt vorhanden
					if (schnittpkt > p.x) {
						// Tausche das Flag
						innen = !innen;
					}
				}
			}
			prev = p1;
		}
		// Wenn prev oberhalb und point unterhalb
		pOberP1Unter = (prev.y >= p.y) && (points.get(0).y < p.y);
		// Wenn prev unterhalb und point oberhalb
		pUnterP1Ober = (prev.y < p.y) && (points.get(0).y >= p.y);
		// Kante zwischen erstem und letztem Punkt des Polygons checken
		if (pOberP1Unter || pUnterP1Ober) {
			schnittpkt = ((new Line(prev, points.get(0)))
					.intersection(p.y, TOP)).x;
			if (p.x <= schnittpkt) {
				innen = !innen;
			}
		}

		// System.out.println(innen);
		return innen;
	}

	/**
	 * Transformiert das Polygon mit Hilfe der übergebenen
	 * Transformationsmatrix.
	 * 
	 * @param m
	 *            die Transformationsmatrix
	 */
	public void transformBy(Matrix m) {
		if (m != null) {
			// Fr jeden Punkt der Polygonliste
			Point newPoint;
			for (Point p : points) {
				// Setze neue Punktkoordinaten
				newPoint = m.multiply(p);
				p.x = newPoint.x;
				p.y = newPoint.y;
			}
		}
	}

	/**
	 * Clippt das Polygon am uebergebenen Rechteck nach der Idee von Sutherland
	 * und Hodgeman. Liefert false, wenn das Polygon nach dem Clipping nicht
	 * mehr der Definition eines Polygons entspricht (mindestens 3-Eck).
	 * 
	 * @param re
	 *            das Clipping-Rechteck
	 * @return true, wenn das Polygon zumindest teilweise im Clippingrechteck
	 *         lag, sonst false
	 */
	public boolean clip(Rectangle re) {
		sutherland_hodgeman(re.getA().x, LEFT); // linke Seite
		sutherland_hodgeman(re.getE().y, BOTTOM); // untere Seite
		sutherland_hodgeman(re.getE().x, RIGHT); // rechte Seite
		sutherland_hodgeman(re.getA().y, TOP); // obere Seite

		// Wenn das Polygon noch teilweise im Clipping-Fenster lag,
		// dann ist es mindestens noch ein Drei-Eck.
		// Ansonsten ist es kein Polygon mehr.
		if (points.size() < 3)
			return false;
		else
			return true;

	}

	/**
	 * Clippt das Polygon an einer Kante des Clipping-Rechtecks.
	 * 
	 * @param value
	 *            der fuer Kante edge interessante Wert in Pixeln
	 * @param edge
	 *            gibt an, ob es sich um eine linke, rechte, obere oder untere
	 *            Kante handelt
	 */
	private void sutherland_hodgeman(int value, int edge) {
		List<Point> tmp = new ArrayList<Point>();

		for (int i = 0; i < points.size(); i++) {
			Point p = points.get(i);
			if (p.onVisibleSide(value, edge))
				tmp.add(p);

			// Prüfe Linie mit dem nächsten Punkt oder mit dem ersten, wenn am
			// Ende
			Line l = new Line(p, (i + 1) < points.size() ? points.get(i + 1)
					: points.get(0));
			Point s = l.intersection(value, edge);

			if (s != null)
				tmp.add(s);
		}

		points = tmp;
	}

	/**
	 * Wendet das Scanline-Verfahren zum Fuellen eines Polygons an
	 * 
	 * @param g
	 *            der grafische Kontext, in den dieses Objekt sich zeichnen soll
	 */
	private void scanlineFill(Graphics g) {
		List<Line> edges = getSortedEdges(); // nach y sort. Kantenliste

		List<Line> activeEdges = new ArrayList<Line>(); // Liste fuer aktive
		// Kanten

		List<Point> intersections; // Liste fuer Schnittpunkte

		boolean notEmpty = true; // akt. Kantenliste leer

		// durchlaufe y-Koord. vom ersten Punkt an, bis aktive Kanten
		// abgearbeitet
		for (int line = edges.get(0).a.y; notEmpty; line--) {
			// besorge absteigend nach y sortierte Liste von aktiven Kanten
			activeEdges = getActiveEdges(edges, activeEdges, line);

			// berechne die Schnittpunkte der akt. Kanten mit der Scanline
			intersections = getSortedIntersections(activeEdges, line);

			// Zeichne im Polygon liegende Teile der Scanline
			paintIntersections(intersections, g);

			notEmpty = !activeEdges.isEmpty();
		}
	}

	/**
	 * Liefert eine Liste, die alle Kanten des Polygons enthaelt. Der y-Wert des
	 * Anfangspunktes einer Kante ist immer groesser, als der y-Wert des
	 * Endpunktes. Waagerechte Kanten werden ignoriert. Die Liste ist absteigend
	 * nach den y-Werten der Anfangspunkte sortiert.
	 * 
	 * @return Kantenliste
	 */
	private List<Line> getSortedEdges() {
		Point prev = null;
		List<Line> edges = new ArrayList<Line>();

		for (Point p : points) {
			if (prev != null && prev.y != p.y) // Waagerechte Linien rauslassen
			{
				if (prev.y > p.y) // Reihenfolge der y-Werte des
					edges.add(new Line(prev, p)); // Anfangs- und Endpunktes
				else
					// beachten
					edges.add(new Line(p, prev));
			}
			prev = p;
		}

		if (prev.y != points.get(0).y) {
			// Polygon schließen (letzte Kante)
			if (prev.y > points.get(0).y) // Reihenfolge der y-Werte des
				edges.add(new Line(prev, points.get(0))); // Anfangs- und
			// Endpunktes
			// beachten

			else
				edges.add(new Line(points.get(0), prev));
		}
		Collections.sort(edges); // Liste sortieren

		return edges;
	}

	/**
	 * Aktualisiere die Liste der aktiven Kanten. Kanten, die in der vorherigen
	 * Scanline geendet sind, werden aus der Liste entfernt, Kanten mit
	 * Startpunkt auf der Scanline werden hinzugefuegt.
	 * 
	 * @param edges
	 *            Kanten des Polygons
	 * @param activeEdges
	 *            Liste der aktiven Kanten
	 * @param line
	 *            Scanline
	 * @return aktualisierte Liste der aktiven Kanten
	 */
	private List<Line> getActiveEdges(List<Line> edges, List<Line> activeEdges,
			int line) {
		for (Line l : edges) {
			if (l.e.y == line)
				activeEdges.remove(l);
			if (l.a.y == line)
				activeEdges.add(l);
		}
		return activeEdges;
	}

	/**
	 * Berechne Schittpunkte der aktiven Kantenliste mit der Scanline und
	 * liefere eine sortierte Liste zurueck.
	 * 
	 * @param activeEdges
	 *            aktive Kanten
	 * @param line
	 *            Scanline
	 * @return sortierte Liste von ggf. doppelten Schnittpunkten
	 */
	private List<Point> getSortedIntersections(List<Line> activeEdges, int line) {
		List<Point> intersections = new ArrayList<Point>();

		for (Line l : activeEdges) {
			int x = (int) (.5 + ((double) l.a.x - l.e.x) / (l.a.y - l.e.y)
					* (line - l.e.y))
					+ l.e.x; // berechne x-Wert
			Point p = new Point(x, line); // erzeuge neuen Punkt

			intersections.add(p);
		} // fuege p sowieso ein

		Collections.sort(intersections); // Schnittpunktmenge nach x sort.

		return intersections;
	}

	/**
	 * Durchlaufe Schnittpunktlinie und zeichne fuer alle Paare von Punkten eine
	 * Linie.
	 * 
	 * @param intersections
	 *            Schnittpunktmenge
	 * @param g
	 *            der grafische Kontext, in den die Linien gezeichnet werden
	 *            sollen
	 */
	private void paintIntersections(List<Point> intersections, Graphics g) {

		Iterator<Point> i = intersections.iterator();

		while (i.hasNext()) {
			Point start = i.next(), end = null;
			if (i.hasNext()) {
				end = i.next();
				Line l = new Line(start, end);
				l.paint(g);
			}
		}
	}
}